/*
 *	ls -lR \B
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "sma02.h"
#include "ps2mc.h"

static int GetFormatVersion(int base);
static int doIt(int base, FILE *fpout);

static const char *FileName = NULL;
//	int mode = ;
static int Mode = LPT_NORMAL;

void Usage(const char *me)
{
	const char *ptr;
	
	ptr = strrchr(me, '\\');
	if (ptr == NULL) ptr = strrchr(me, ':');
	if (ptr == NULL) ptr = me;
	else ptr ++;
	
	fprintf(stderr, "%s Build:%s. Programmed by Hanimar.\n", ptr, __DATE__);
	fprintf(stderr, "Usage: %s [/b] [/o<filename to output>]\n", ptr);
	fputs("/b : Bidirectional mode.\n", stderr);
	exit(EXIT_FAILURE);
}
void argchk(int argc, char **argv)
{
	int i, j;
	int prev = 0;
	
	for (i = 1 ; i < argc ; i ++) {
		if (argv[i][0] == '/' || argv[i][0] == '-') {
			for (j = 1 ; argv[i][j] != '\0' ; j ++) {
				switch (argv[i][j]) {
				case 'b':
				case 'B':
					Mode = LPT_BIDIRECTIONAL;
					break;
				case 'o':
				case 'O':
					if (argv[i][j + 1] != '\0') {
						FileName = &argv[i][2];
						while (argv[i][j + 1] != '\0') j ++;
					} else prev = 'o';
					break;
				default:
					Usage(argv[0]);
				}
			}
			continue;
		}
		switch (prev) {
		case 'o':
			FileName = argv[i];
			break;
		default:
			Usage(argv[0]);
		}
		prev = 0;
	}
	if (prev == 'o' && FileName == NULL) Usage(argv[0]);
	
	return;
}

int main(int argc, char **argv)
{
	int base;
	UCHAR status;
	int version;
	int rc;
	int retry;
	FILE *fpout = NULL;
	
	argchk(argc, argv);
	
	if (FileName != NULL) {
		fpout = fopen(FileName, "w");
		if (fpout == NULL) {
			fprintf(stderr, "Can not open file \"%s\".\n", FileName);
			exit(EXIT_FAILURE);
		}
	}
	
	base = LptGetPortAddr(1);	/* LPT1 = 378H ̂͂ */
	rc = Sma02Open(base, Mode);
	if (rc < 0) {
		Sma02Close(base);
		WaitMSec(200L);
		rc = Sma02Open(base, Mode);
		if (rc < 0) {
			fputs("Can not connect to SmartMedia via LPT1.\n", stderr);
			Sma02Close(base);
			return EXIT_FAILURE;
		}
	}
	/* 200 ~bقǑ҂܂BӖ͂ȂłǁB */
	WaitMSec(200L);
	Sma02Write1Byte(base, 1, 0x1c);	/* VCC ON */
	WaitMSec(20L);

	/* fBAĂ邩ׂ܂ */
	for (retry = 100 ; retry ; retry --) {
		status = Sma02Read1Byte(base, 1);
		if ((status & 0x80) == 0) break;
		Sma02Reset(base);
	}
	if ((status & 0x04) == 0) {
		fprintf(stderr, "Make sure SmartMedia is set correctly.(%02Xh)\n", status);
		Sma02Write1Byte(base, 1, 0x18);	/* VCC OFF */
		Sma02Close(base);
		return EXIT_FAILURE;
	}
	
	/* PlayStation2 tH[}bgǂׂ܂傤B */
	if ((version = GetFormatVersion(base)) < 0) {
		fputs("This can not be a PlayStation2 format.\n", stderr);
		Sma02Write1Byte(base, 1, 0x18);	/* VCC OFF */
		Sma02Close(base);
		return EXIT_FAILURE;
	}
	if (version != 1100) {
		fprintf(stderr, "Warning:Unknown version: %d\n", version);
	}

	doIt(base, fpout != NULL ? fpout:stdout);
	if (fpout != NULL) fclose(fpout);
	
	Sma02Close(base);
	return EXIT_SUCCESS;
}

/*
 +	PlayStation2 tH[}bgǂׂB
 +	iʐMeXg˂ĂAȁj
 +	1 sector ڂ̈ꔭڂ
 +	"Sony PS2 Memory Card Format 1.1.0.0" Ă̂܂B
 +	1.1.0.0 ̂ƂǂŁA 1100 ԂA݂ɂĂ܂B
 */
static int GetFormatVersion(int base)
{
	UCHAR buf[528];
	FILE *fp;
	UCHAR *p;
	int ret = 0;
	
	Sma02ReadPage(base, buf, 0, 1, NULL, NULL);
	if (memcmp(&buf[9], "Memory Card", 11) == 0) {
		printf("Volume is \"%s\".\n\n", buf);
		for (p = &buf[16] ; *p ; p ++) {
			if (isdigit(*p)) {
				ret *= 10;
				ret += *p - '0';
			}
		}
	} else return -1;
	
/*********
	fp = fopen("c:\\tmp\\a.out", "wb");
	fwrite(buf, 1, 528, fp);
	fclose(fp);
**********/
	return ret;
}

/*
 *	ls -lR ̖{
 *	܂fBNg̒gS\āA
 *	ɂ̉̃fBNgǂށAċABvHAcnrB
 *	X^bNڂɐݒ肵ƂȁB
 */
static int lslr(int dirFatNo);

static int Base;
static PS2MCFAT Fat;
static UCHAR *Sector;
static FILE *FpOut;

static int doIt(int base, FILE *fpout)
{
	int root;
	int rc;
	
#define FATDATASECTOR	0x29

	Sector = (UCHAR *)malloc(1056);	/* (512+16)*2 */
	if (Sector == NULL) {
		fprintf(stderr, "Can not allocate memory (%d)\n", 1056);
		return -1;
	}
	if (Ps2mcFatInitialize(&Fat, 4) < 0) {
		fprintf(stderr, "Can not allocate memory (%d)\n", sizeof (struct ps2mcFatTable_t));
		free(Sector);
		return -1;
	}
	
	/* [gfBNg̃ZN^ԍ擾܂B 0+34H */
	root = Ps2mcGetRootSectorNo(base);
	if (root < FATDATASECTOR || root > 8192) {
		fprintf(stderr, "Invalid root sector %d.\n", root);

		Ps2mcFatDispose(&Fat);
		free(Sector);
		return -2;
	}
	root -= FATDATASECTOR;

	Base = base;
	FpOut = fpout;
	
	rc = lslr(root);
	
	Ps2mcFatDispose(&Fat);
	free(Sector);

	return rc;
}

/*
 *	BIG-ENDIAN vbgtH[΍BœCH
 *	LITTLE-ENDIAN  32bit lǂ݂܂B
 */
static unsigned long readInt32(void *ptr)
{
	unsigned char *p = (unsigned char *)ptr;
	unsigned long ret = 0;
	int i;
	
	for (i = 4 ; i ; i --) {
		ret <<= 8;
		ret |= p[i - 1];
	}
	return ret;
}
static unsigned int readInt16(void *ptr)
{
	unsigned char *p = (unsigned char *)ptr;
	unsigned long ret = 0;
	int i;
	
	for (i = 2 ; i ; i --) {
		ret <<= 8;
		ret |= p[i - 1];
	}
	return ret;
}

/*
 *	ċAŌĂ΂܂B
 *	X^bN̂ static ϐg܂B
 */

static int lslr(int dirFatNo)
{
	int end = 0;
	int i;
	int fatNo = dirFatNo;
	/* t@C̗ */
	UCHAR *dirPtr, *p;
	int ret;
	
	for (fatNo = dirFatNo ; fatNo >= 0 ; fatNo = Ps2mcFatGetNext(&Fat, Base, fatNo)) {
		Sma02ReadSector(Base, Sector, (fatNo + FATDATASECTOR) << 1);
		/* P̃fBNg 512 bytes Ȃ̂łQ */
		for (i = 0 ; i < 2 ; i ++) {
			dirPtr = Sector;
			if (i) dirPtr += 512;
			
			if (readInt16(dirPtr) == 0xffff) {	/* I肾B */
				end ++;
				break;
			}
			/*  +00h */
			/* t@CTCY +04h */
			fprintf(FpOut, "%02X%02X %7ld ", dirPtr[0], dirPtr[1], readInt32(dirPtr + 4));
			/* XVt +19H YYYY-MM-DD HH:MM:SS Ȃ琢EW낤B */
			p = dirPtr + 0x19;
			fprintf(FpOut, "%04d-%02d-%02d %02d:%02d:%02d ",
				readInt16(p + 5), p[4], p[3], p[2], p[1], p[0]);
			/* t@C +40h */
			fprintf(FpOut, "%s\n", dirPtr + 0x40);
		}
		if (end) break;
	}
	/*
	 *	xǂŁAfBNg@Ă܂B
	 *	X^bNӁI[ȂƁB
	 */
	end = 0;
	for (fatNo = dirFatNo ; fatNo >= 0 ; fatNo = Ps2mcFatGetNext(&Fat, Base, fatNo)) {
		/* P̃fBNg 512 bytes Ȃ̂łQ */
		for (i = 0 ; i < 2 ; i ++) {
			/* ċA Sector ̂ŁAǂށA */
			Sma02ReadSector(Base, Sector, (fatNo + FATDATASECTOR) << 1);
			dirPtr = Sector;
			if (i) dirPtr += 512;
			
			if (readInt16(dirPtr) == 0xffff) {	/* I肾B */
				end ++;
				break;
			}
//			fprintf(FpOut, "%s=%02X\n", dirPtr + 0x40, *dirPtr);
			if (!(dirPtr[0] & 0x20)) continue;
			/* fBNg */
			/* "."  ".." ͖ */
			if (!strcmp(dirPtr + 0x40, ".")) continue;
			if (!strcmp(dirPtr + 0x40, "..")) continue;
			/* fBNg\ */
			fprintf(FpOut, "\n%s:\n\n", dirPtr + 0x40);
			
			ret = lslr(readInt16(dirPtr + 0x10));
			if (ret != 0) return ret;
		}
		if (end) break;
	}
	
	return 0;
	
}
