/*
 *	TABSTOP=4
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "sma02.h"
#include "ps2mc.h"

/* Is End Of Chain */
#define ISEOC(fat, tableNo) ( ((tableNo) < 0) || ((fat)->numSector <= tableNo) )
/*
 *	FAT e[uǂނ߂̏
 *	sectorSize ̓ɊmۂZN^B
 *	Pȏw肵ĂB20ƑSǂ߂܂AB
 *	ȂꍇŌɓǂ񂾃ZN^C[Wy[WAEg
 *	̂ŏȂĂ邢łB
 *
 *	͊mۂZN^
 *	s͂OȉB
 */
int Ps2mcFatInitialize(PS2MCFAT *fat, int numSector)
{
	int i;
	struct ps2mcFatTable_t *tables;
	struct ps2mcFatTable_t *table;
	
	if (numSector <= 0) return 0;
	do {
		tables = (struct ps2mcFatTable_t *)malloc(
			sizeof (struct ps2mcFatTable_t) * numSector
			);
		if (tables != NULL) break;
	} while (-- numSector > 0);
	if (numSector == 0) return -1;
	
	fat->tables = tables;
	fat->numSector = numSector;
	fat->pri_top = -1;
	fat->pri_bottom = 0;
	
	for (i = 0 ; i < numSector ; i ++) {
		table = &tables[i];
		
		table->sectorNo = -1;	/* J */
		/* I[͔͈͊O̒lƂ܂B */
		table->pri_prev = i + 1;
		table->pri_next = i - 1;
	}

	return numSector;
}

void Ps2mcFatDispose(PS2MCFAT *fat)
{
	free(fat->tables);
	/* ȂƂȂBꂢɂ܂B */
	fat->tables = NULL;
	fat->numSector = 0;
	
	return;
}

/*
 *	BIG-ENDIAN vbgtH[΍B
 *	LITTLE-ENDIAN  32bit lǂ݂܂B
 */
static unsigned long readInt32(void *ptr)
{
	unsigned char *p = (unsigned char *)ptr;
	unsigned long ret = 0;
	int i;
	
	for (i = 4 ; i ; i --) {
		ret <<= 8;
		ret |= p[i - 1];
	}
	return ret;
}

/*
 *	fatNo ɑ FAT# ԂB
 *	߂l
 *	>0 
 *	-1: fatNo ͏I[B
 *	-2: fatNo ͂܂ƂłȂB
 */
int Ps2mcFatGetNext(PS2MCFAT *fat, int iobase, int fatNo)
{
	int tableNo;
	int destSectorNo;
	struct ps2mcFatTable_t *table;
	unsigned long next;
	
	/* 1sector  256 ̃Gg܂B */
	destSectorNo = fatNo >> 8;
	for (tableNo = fat->pri_top ; !ISEOC(fat, tableNo) ; tableNo = table->pri_next) {
		table = &fat->tables[tableNo];
//		printf("Searching sector#%d: tables[%d].sectorNo=%d\n", destSectorNo, tableNo, table->sectorNo);
		if (table->sectorNo == destSectorNo) break;
	}
	if (ISEOC(fat, tableNo)) {
		/* ɂȂ̂œǂݍށB */
		tableNo = fat->pri_bottom;
		table = &fat->tables[tableNo];
		Sma02ReadSector(
			iobase,						/* LPT  Base Address */
			table->table,						/* Page Data  buffer */
			0x12 + (destSectorNo << 1)	/* ǂݍރy[Wԍ */
			);
		table->sectorNo = destSectorNo;

		/* Xg͂ */
		fat->pri_bottom = table->pri_prev;
		fat->tables[table->pri_prev].pri_next = -1;
		/* Xg̐擪ɎĂ */
		table->pri_prev = -1;
		table->pri_next = fat->pri_top;
		if (!ISEOC(fat, fat->pri_top)) fat->tables[fat->pri_top].pri_prev = tableNo;
		fat->pri_top = tableNo;
	} else if (tableNo != fat->pri_top) {
		/* ς胊Xg̐擪ɂĂ */
		/* Xg͂ */
		fat->tables[table->pri_prev].pri_next = table->pri_next;
		if (!ISEOC(fat, table->pri_next))
			fat->tables[table->pri_next].pri_prev = table->pri_prev;
		/* Xg̐擪ɎĂ */
		table->pri_prev = -1;
		table->pri_next = fat->pri_top;
		if (!ISEOC(fat, fat->pri_top)) fat->tables[fat->pri_top].pri_prev = tableNo;
		fat->pri_top = tableNo;
	}
	/*  tableNo  table ob`B */
	next = readInt32(&table->table[(fatNo & 0xff) * 4]);

	if (next == 0xffffffff) return -1;
	if (!(next & 0x80000000)) return -2;
	return next & 0x7fff;
}

/*
 *	root fBNg̊i[ꂽZN^ԍԂBB
 *	ɂ 29h Ă񂾂ǁAӖႤB
 *	\zł́A
 *	FAT ̃f[^ 29H n܂̂͌ŒB
 *	 34H ̏ꏊɂroot ̃fBNgGg̏ꏊɏB
 *	Ȃ񂾂ǂˁB
 */
int Ps2mcGetRootSectorNo(int iobase)
{
	UCHAR buf[528];
	
	Sma02ReadPage(iobase, buf, 0, 1, NULL, NULL);
	return readInt32(&buf[0x34]);

}
